# -*- coding: utf-8 -*-
from distutils.core import setup

modules = \
['pygenstub']
install_requires = \
['docutils>=0.14,<0.15']

entry_points = \
{'console_scripts': ['pygenstub = pygenstub:main']}

setup_kwargs = {
    'name': 'pygenstub',
    'version': '1.2.3',
    'description': 'Python stub file generator.',
    'long_description': 'Copyright (C) 2016-2019 H. Turgut Uyar <uyar@tekir.org>\n\npygenstub is a utility for generating stub files from docstrings\nin source files. It takes a source file as input and creates a stub file\nwith the same base name and the ``.pyi`` extension.\n\n:PyPI: https://pypi.org/project/pygenstub/\n:Repository: https://github.com/uyar/pygenstub\n:Documentation: https://pygenstub.readthedocs.io/\n\nIf the docstring of a function includes a **sig** field, the value of that\nfield will be used to generate a prototype by matching the types to the\nparameters in the same order. For example, for the code given below:\n\n.. code-block:: python\n\n   def foo(a, b):\n       """Do foo.\n\n       :sig: (int, str) -> None\n       """\n\npygenstub will generate the following stub:\n\n.. code-block:: python\n\n   def foo(a: int, b: str) -> None: ...\n\npygenstub consists of a single module which itself contains signatures.\nYou can see the `source code`_ and the autogenerated `stub file`_\nas an example.\n\nUsage\n-----\n\npygenstub has been tested with Python 2.7, Python 3.4+, and compatible\nversions of PyPy. You can install the latest version using ``pip``::\n\n  pip install pygenstub\n\nInstallation creates a script named ``pygenstub`` which can be used\nas follows::\n\n  pygenstub foo.py\n\nThis command will generate the file ``foo.pyi`` in the same directory\nas the input file. If the output file already exists, it will be overwritten.\n\nSphinx autodoc support\n----------------------\n\nIf pygenstub is activated as a Sphinx extension (after autodoc), it will insert\ntype comments for parameters and return values in the docstring. It will also\nremove the signature fields to exclude them from the output:\n\n.. code-block:: python\n\n   extensions = [\n       \'sphinx.ext.autodoc\',\n       \'pygenstub\'\n   ]\n\nAs an example of the output, you can see the `API documentation`_\nfor pygenstub itself.\n\n.. _source code: https://github.com/uyar/pygenstub/blob/master/pygenstub.py\n.. _stub file: https://github.com/uyar/pygenstub/blob/master/pygenstub.pyi\n.. _API documentation: https://pygenstub.readthedocs.io/en/latest/api.html\n',
    'author': 'H. Turgut Uyar',
    'author_email': 'uyar@tekir.org',
    'url': 'https://github.com/uyar/pygenstub',
    'py_modules': modules,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*',
}


setup(**setup_kwargs)
